/**
 * \file L2.cpp
 *
 *  \date Jan 22, 2013
 *  \author tuleu
 */

#include "L2.h"

#include <liboncilla-webots/Common.h>
#include <webots/Servo.hpp>
#include <liboncilla-webots/Config.h>

#include <libwebots/communication/client.hh>
#include <libwebots/plugin/modules/jointmanager/jointmanagerbuilder.hh>



namespace liboncilla {
namespace webots {


L2::L2(rci::oncilla::Synchronizer & s,
       const std::string & name,
       uint32_t identifier,
       ::webots::Servo * servoInline,
        ::webots::Servo * servoDiagonal,
       const config::LegSection & config)
	: rci::oncilla::L2(s,name)
	, d_servoInline(*servoInline)
	, d_servoDiagonal(*servoDiagonal)
	, d_joint(identifier)
	, d_minPosition(config.Diagonal().MinLength())
	, d_maxPosition(config.Diagonal().MaxLength())
	, d_maxSpeed(std::fabs(config.MaxSpeed())){
	if(!servoInline || !servoDiagonal){
		throw std::out_of_range("Could not found suitable webots Servo for '" + name + "'");
	}

	Init(config);
}

L2::~L2(){
}

void L2::Activate(){
	d_servoInline.enablePosition(1);
	d_servoDiagonal.enablePosition(1);
	d_passes = 0;

};

void L2::Deactivate(){
	d_servoInline.disablePosition();
	d_servoDiagonal.disablePosition();
};

void L2::PreStep(double ts){
	using namespace libwebots::communication;
	using namespace libwebots::communication;
	namespace lm=libwebots::plugin::modules::messages;


	if(d_passes == 0){
		d_lastCommand = d_minPosition;
		++d_passes;
	}

	rci::ControlModePtr c = getControlMode();

	if(c->supports(rci::jointPositionControl) ){
		using namespace lm::builders;
		//get the command and apply joint limit position
		double command = clamp(getLastPositionCommand()->rad(0),0.0,1.0);
		command *= d_maxPosition - d_minPosition;
		command += d_minPosition;

		//now apply speed limitation
		double commandRate = command - d_lastCommand;
		commandRate /= ts;

		if(std::abs(commandRate) > d_maxSpeed){
			command = d_lastCommand + ts * d_maxSpeed * (commandRate > 0 ? 1.0 : -1.0);
			//still inside the position boundaries by construction
		}

		Client::Instance().Send(JointManager::SetJointParameter(d_joint,
		                                                        lm::jointmanager::JointParameters::dParamLoStop,
		                                                        command));
		d_lastCommand = command;
	} else {
		throw std::runtime_error("Please use jointPositionControl for L2.");
	}
}

void L2::PostStep(double ts){
	double pos = d_servoDiagonal.getPosition() - d_minPosition;
	pos /= d_maxPosition - d_minPosition;
	_latestJointPosition->setFromRad(rci::oncilla::L2::MotorAxisIndex,
	                                 pos);
	_latestJointPosition->setFromRad(rci::oncilla::L2::MagneticEncoderIndex,
	                                 d_servoInline.getPosition());

}


void L2::Init(const config::LegSection & config){
	using namespace libwebots::communication;
	using namespace libwebots::communication;
	namespace lm=libwebots::plugin::modules::messages;

	Client & c = Client::Instance();

	using namespace lm::builders;


	JointManager::JointParameterValue * p;
	JointManager::ListOfParameterValue pv;
	pv.Reserve(4);

	p = pv.Add();
	double margin = config.Margin() * (d_maxPosition - d_minPosition);
	p->set_parameter(lm::jointmanager::JointParameters::dParamHiStop);
	p->set_value(d_maxPosition + margin);

	p = pv.Add();
	p->set_parameter(lm::jointmanager::JointParameters::dParamLoStop);
	p->set_value(d_minPosition);

	p = pv.Add();
	p->set_parameter(lm::jointmanager::JointParameters::dParamStopERP);
	p->set_value(config.DiagonalErp());

	p = pv.Add();
	p->set_parameter(lm::jointmanager::JointParameters::dParamStopCFM);
	p->set_value(config.DiagonalCfm());

	c.Send(JointManager::SetJointParameters(d_joint,pv));

}

} /* namespace webots */
} /* namespace liboncilla */
